@@ -16,12 +16,15 @@
 
 from rax import types
 from rax import utils
+from rax._src.losses import listmle_loss
 from rax._src.losses import pairwise_hinge_loss
 from rax._src.losses import pairwise_logistic_loss
 from rax._src.losses import pairwise_mse_loss
 from rax._src.losses import pointwise_mse_loss
 from rax._src.losses import pointwise_sigmoid_loss
+from rax._src.losses import poly1_softmax_loss
 from rax._src.losses import softmax_loss
+from rax._src.losses import unique_softmax_loss
 from rax._src.metrics import ap_metric
 from rax._src.metrics import dcg_metric
 from rax._src.metrics import mrr_metric
@@ -32,15 +35,18 @@ from rax._src.t12n import approx_t12n
 from rax._src.t12n import bound_t12n
 from rax._src.t12n import gumbel_t12n
 
-__version__ = "0.1.0"
+__version__ = "0.2.0"
 
 # pyformat: disable
 __all__ = [
+    "listmle_loss",
     "pairwise_hinge_loss",
     "pairwise_logistic_loss",
     "pairwise_mse_loss",
     "pointwise_mse_loss",
     "pointwise_sigmoid_loss",
+    "poly1_softmax_loss",
+    "unique_softmax_loss",
     "softmax_loss",
     "ap_metric",
     "dcg_metric",
