@@ -58,6 +58,7 @@ from typing import Callable, Optional
 import jax
 import jax.numpy as jnp
 
+from rax._src import metrics
 from rax._src import utils
 from rax._src.types import Array
 from rax._src.types import ReduceFn
@@ -69,14 +70,14 @@ def softmax_loss(scores: Array,
                  where: Optional[Array] = None,
                  weights: Optional[Array] = None,
                  label_fn: Callable[..., Array] = lambda a, where: a,
-                 reduce_fn: Optional[ReduceFn] = jnp.sum) -> Array:
+                 reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
   r"""Softmax loss.
 
   Definition:
 
   .. math::
       \ell(s, y) =
-      \sum_i y_i \log \frac{\exp(s_i)}{\sum_j \exp(s_j)}
+      - \sum_i y_i \log \frac{\exp(s_i)}{\sum_j \exp(s_j)}
 
   Args:
     scores: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
@@ -97,7 +98,7 @@ def softmax_loss(scores: Array,
   Returns:
     The softmax loss.
   """
-  # Applies where so that whereed elements do not count towards the loss.
+  # Applies mask so that masked elements do not count towards the loss.
   if where is not None:
     labels = jnp.where(where, labels, jnp.zeros_like(labels))
     scores = jnp.where(where, scores, -jnp.ones_like(scores) * jnp.inf)
@@ -118,6 +119,209 @@ def softmax_loss(scores: Array,
   return utils.safe_reduce(loss, reduce_fn=reduce_fn)
 
 
+def poly1_softmax_loss(scores: Array,
+                       labels: Array,
+                       *,
+                       epsilon: float = 1.0,
+                       where: Optional[Array] = None,
+                       weights: Optional[Array] = None,
+                       reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
+  r"""Poly1 softmax loss.
+
+  Definition :cite:p:`leng2022polyloss`:
+
+  .. math::
+      \ell(s, y) = softmax(s, y) + \epsilon * (1 - pt)
+
+  where :math:`softmax` is the standard softmax loss as implemented in
+  :func:`~rax.softmax_loss` and :math:`pt` is the target softmax probability
+  defined as:
+
+  .. math::
+      pt = \sum_i \frac{y_i}{\sum_j y_j} \frac{\exp(s_i)}{\sum_j \exp(s_j)}
+
+  Args:
+    scores: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
+      score of each item.
+    labels: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
+      relevance label for each item.
+    epsilon: A float hyperparameter indicating the weight of the leading
+      polynomial coefficient in the poly loss.
+    where: An optional ``[..., list_size]``-:class:`~jax.numpy.ndarray`,
+      indicating which items are valid for computing the loss. Items for which
+      this is False will be ignored when computing the loss.
+    weights: An optional ``[..., list_size]``-:class:`~jax.numpy.ndarray`,
+      indicating the weight for each item.
+    reduce_fn: An optional function that reduces the loss values. Can be
+      :func:`jax.numpy.sum` or :func:`jax.numpy.mean`. If ``None``, no reduction
+      is performed.
+
+  Returns:
+    The poly1 softmax loss.
+  """
+  # Compute softmax cross-entropy loss without batch reduction.
+  ce = softmax_loss(
+      scores, labels, where=where, weights=weights, reduce_fn=None)
+
+  # Applies mask so that masked elements do not count towards the loss.
+  if where is not None:
+    labels = jnp.where(where, labels, jnp.zeros_like(labels))
+    scores = jnp.where(where, scores, -jnp.ones_like(scores) * jnp.inf)
+
+  # Apply weights to labels.
+  if weights is not None:
+    labels *= weights
+
+  # Compute target probabilities.
+  scores_softmax = jax.nn.softmax(scores)
+  labels_normalized = utils.normalize_probabilities(labels, where=where)
+  pt = jnp.sum(labels_normalized * scores_softmax, where=where, axis=-1)
+
+  # For lists where all items are masked, this sets pt to 1 so that the term
+  # (1 - pt) is set to 0 for the loss computation.
+  if where is not None:
+    pt = jnp.where(jnp.all(jnp.logical_not(where), axis=-1), 1., pt)
+
+  # Compute and return the poly1 loss.
+  loss = ce + epsilon * (1. - pt)
+  return utils.safe_reduce(loss, reduce_fn=reduce_fn)
+
+
+def unique_softmax_loss(scores: Array,
+                        labels: Array,
+                        *,
+                        where: Optional[Array] = None,
+                        weights: Optional[Array] = None,
+                        gain_fn: Optional[Callable[
+                            [Array], Array]] = metrics.default_gain_fn,
+                        reduce_fn: ReduceFn = jnp.mean) -> Array:
+  r"""Unique softmax loss.
+
+  Definition :cite:p:`zhu2020listwise`:
+
+  .. math::
+      \ell(s, y) =
+      - \sum_i \operatorname{gain}(y_i)
+      \log \frac{\exp(s_i)}{\exp(s_i) + \sum_{j : y_j < y_i} \exp(s_j)}
+
+  where :math:`\operatorname{gain}(y_i)` is a user-specified gain function
+  applied to label :math:`y_i` to boost items with higher relevance.
+
+  Args:
+    scores: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
+      score of each item.
+    labels: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
+      relevance label for each item.
+    where: An optional ``[..., list_size]``-:class:`~jax.numpy.ndarray`,
+      indicating which items are valid for computing the loss. Items for which
+      this is False will be ignored when computing the loss.
+    weights: An optional ``[..., list_size]``-:class:`~jax.numpy.ndarray`,
+      indicating the weight for each item.
+    gain_fn: An optional function that maps relevance labels to gain values. If
+      provided, the per-item losses are multiplied by ``gain_fn(label)`` to
+      boost the importance of relevant items.
+    reduce_fn: An optional function that reduces the loss values. Can be
+      :func:`jax.numpy.sum` or :func:`jax.numpy.mean`. If ``None``, no reduction
+      is performed.
+
+  Returns:
+    The unique softmax loss.
+  """
+  # Construct pairwise matrices for scores and labels. The labels matrix will
+  # indicate, for each item, which other items have a smaller label.
+  labels_lt = jnp.expand_dims(labels, -2) < jnp.expand_dims(labels, -1)
+  scores_repeated = jnp.repeat(
+      jnp.expand_dims(scores, -2), scores.shape[-1], axis=-2)
+
+  # Build an identity mask to select items during softmax computation.
+  identity_mask = jnp.identity(scores.shape[-1], dtype=jnp.bool_)
+
+  # Apply mask to ignore invalid items.
+  if where is not None:
+    labels_lt &= jnp.expand_dims(where, -2)
+    identity_mask &= jnp.expand_dims(where, -2)
+
+  # Compute log-softmax for each of the items' unique softmax distribution. This
+  # effectively computes the log_softmax for each item using only itself and
+  # items with a smaller relevance label in the denominator. The computed
+  # log_softmax distribution is reduced by selecting its diagonal elements.
+  log_softmax = jax.nn.log_softmax(
+      scores_repeated,
+      axis=-1,
+      where=(identity_mask | labels_lt),
+      initial=jnp.min(scores))
+  log_softmax = jnp.diagonal(log_softmax, axis1=-2, axis2=-1)
+
+  # Apply per-item weights.
+  if weights is not None:
+    log_softmax *= weights
+
+  # Apply per-item `gain_fn` boost.
+  if gain_fn is not None:
+    log_softmax *= gain_fn(labels)
+
+  # Compute per-list loss and return a reduced loss.
+  loss = -jnp.sum(log_softmax, axis=-1, where=where)
+  return utils.safe_reduce(loss, reduce_fn=reduce_fn)
+
+
+def listmle_loss(scores: Array,
+                 labels: Array,
+                 *,
+                 key: Optional[Array] = None,
+                 where: Optional[Array] = None,
+                 reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
+  r"""ListMLE Loss.
+
+  .. note::
+
+    This loss performs sorting using the given labels. If the labels contain
+    multiple identical values, you should provide a :func:`~jax.random.PRNGKey`
+    to the ``key`` argument to make sure ties are broken randomly during the
+    sorting operation.
+
+  Definition :cite:p:`xia2008listwise`:
+
+  .. math::
+      \ell(s, y) =
+      - \sum_i \log
+        \frac{\exp(s_i)}{\sum_j I[rank(y_j) \ge rank(y_i)] \exp(s_j)}
+
+  where :math:`\operatorname{rank}(y_i)` indicates the rank of item :math:`i`
+  after sorting all labels :math:`y`.
+
+  Args:
+    scores: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
+      score of each item.
+    labels: A ``[..., list_size]``-:class:`~jax.numpy.ndarray`, indicating the
+      relevance label for each item.
+    key: An optional :func:`~jax.random.PRNGKey` to perform random tie-breaking.
+    where: An optional ``[..., list_size]``-:class:`~jax.numpy.ndarray`,
+      indicating which items are valid for computing the loss. Items for which
+      this is False will be ignored when computing the loss.
+    reduce_fn: An optional function that reduces the loss values. Can be
+      :func:`jax.numpy.sum` or :func:`jax.numpy.mean`. If ``None``, no reduction
+      is performed.
+
+  Returns:
+    The listmle loss.
+  """
+  # Sort scores and mask by labels.
+  if where is None:
+    where = jnp.ones_like(scores, dtype=jnp.bool_)
+
+  scores_sorted, where_sorted = utils.sort_by(
+      labels, [scores, where], where=where, key=key)
+
+  # Compute cumulative logsumexp.
+  lse = utils.logcumsumexp(
+      scores_sorted, axis=-1, where=where_sorted, reverse=True)
+
+  # Reduce list MLE loss.
+  loss = -jnp.sum(scores_sorted - lse, axis=-1, where=where_sorted)
+  return utils.safe_reduce(loss, reduce_fn=reduce_fn)
+
+
 def compute_pairs(a: Array, op: Callable[[Array, Array], Array]) -> Array:
   """Computes pairs based on values of `a` and the given pairwise `op`.
 
@@ -142,7 +346,7 @@ def pairwise_hinge_loss(scores: Array,
                         *,
                         where: Optional[Array] = None,
                         weights: Optional[Array] = None,
-                        reduce_fn: ReduceFn = jnp.sum) -> Array:
+                        reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
   r"""Pairwise hinge loss.
 
   Definition:
@@ -193,7 +397,7 @@ def pairwise_logistic_loss(scores: Array,
                            *,
                            where: Optional[Array] = None,
                            weights: Optional[Array] = None,
-                           reduce_fn: ReduceFn = jnp.sum) -> Array:
+                           reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
   r"""Pairwise logistic loss.
 
   Definition :cite:p:`burges2005learning`:
@@ -244,7 +448,7 @@ def pointwise_sigmoid_loss(scores: Array,
                            *,
                            where: Optional[Array] = None,
                            weights: Optional[Array] = None,
-                           reduce_fn: ReduceFn = jnp.sum) -> Array:
+                           reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
   r"""Sigmoid cross entropy loss.
 
   Definition:
@@ -293,7 +497,7 @@ def pointwise_mse_loss(scores: Array,
                        *,
                        where: Optional[Array] = None,
                        weights: Optional[Array] = None,
-                       reduce_fn: ReduceFn = jnp.sum) -> Array:
+                       reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
   r"""Mean squared error loss.
 
   Definition:
@@ -331,7 +535,7 @@ def pairwise_mse_loss(scores: Array,
                       *,
                       where: Optional[Array] = None,
                       weights: Optional[Array] = None,
-                      reduce_fn: ReduceFn = jnp.sum) -> Array:
+                      reduce_fn: Optional[ReduceFn] = jnp.mean) -> Array:
   r"""Pairwise mean squared error loss.
 
   Definition:
