@@ -16,12 +16,13 @@
 """Tests for rax._src.losses."""
 
 import doctest
+import functools
 import math
 from absl.testing import absltest
 from absl.testing import parameterized
 import jax
-from jax import test_util as jtu
 import jax.numpy as jnp
+import numpy as np
 
 import rax
 from rax._src import losses
@@ -33,7 +34,7 @@ logloss = lambda x: log(1. + exp(-x))
 sigmoid = lambda x: 1. / (1. + exp(-x))
 
 
-class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
+class LossesTest(parameterized.TestCase):
 
   @parameterized.parameters([{
       "loss_fn":
@@ -41,6 +42,38 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
       "expected_value":
           -(log(exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
             log(exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))))
+  }, {
+      "loss_fn":
+          losses.listmle_loss,
+      "expected_value":
+          -sum([
+              log(exp(1.) / (exp(1.) + exp(2.) + exp(0.) + exp(3.))),
+              log(exp(2.) / (exp(2.) + exp(0.) + exp(3.))),
+              log(exp(0.) / (exp(0.) + exp(3.))),
+              log(exp(3.) / (exp(3.))),
+          ])
+  }, {
+      "loss_fn":
+          losses.poly1_softmax_loss,
+      "expected_value":
+          -(log(exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))) +
+          (1. - (0.5 * exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)) + 0.5 *
+                 (exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))))
+  }, {
+      "loss_fn":
+          losses.unique_softmax_loss,
+      "expected_value":
+          -(log(exp(2.) / (exp(0.) + exp(3.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.))))
+  }, {
+      "loss_fn":
+          functools.partial(losses.poly1_softmax_loss, epsilon=0.1),
+      "expected_value":
+          -(log(exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))) + 0.1 *
+          (1. - (0.5 * exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)) + 0.5 *
+                 (exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))))
   }, {
       "loss_fn": losses.pairwise_hinge_loss,
       "expected_value": (3. - 1. + 1.) + (3. - 2. + 1.)
@@ -76,9 +109,9 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     scores = jnp.asarray([0., 3., 1., 2.])
     labels = jnp.asarray([0., 0., 1., 1.])
 
-    loss = loss_fn(scores, labels)
+    loss = loss_fn(scores, labels, reduce_fn=jnp.sum)
 
-    self.assertArraysAllClose(jnp.asarray(expected_value), loss)
+    np.testing.assert_allclose(jnp.asarray(expected_value), loss)
 
   @parameterized.parameters([{
       "loss_fn":
@@ -86,6 +119,21 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
       "expected_value":
           -((-2.1e26 - (0. + -2.1e26 + 3.4e37 + 42.)) +
             (3.4e37 - (0. + -2.1e26 + 3.4e37 + 42.)))
+  }, {
+      "loss_fn": losses.listmle_loss,
+      "expected_value": 3.4e37
+  }, {
+      "loss_fn":
+          losses.poly1_softmax_loss,
+      "expected_value":
+          -((-2.1e26 - (0. + -2.1e26 + 3.4e37 + 42.)) +
+            (3.4e37 - (0. + -2.1e26 + 3.4e37 + 42.)))
+  }, {
+      "loss_fn":
+          losses.unique_softmax_loss,
+      "expected_value":
+          -((-2.1e26 - (0. + 42.)) +
+            (3.4e37 - (0. + 3.4e37 + 42.)))
   }, {
       "loss_fn": losses.pairwise_hinge_loss,
       "expected_value": (1. - (-2.1e26 - 0.)) + (1. - (-2.1e26 - 42.0))
@@ -114,13 +162,35 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     scores = jnp.asarray([0., -2.1e26, 3.4e37, 42.0])
     labels = jnp.asarray([0., 1., 1., 0.])
 
-    loss = loss_fn(scores, labels)
+    loss = loss_fn(scores, labels, reduce_fn=jnp.sum)
 
-    self.assertArraysAllClose(jnp.asarray(expected_value), loss)
+    np.testing.assert_allclose(jnp.asarray(expected_value), loss)
 
   @parameterized.parameters([{
+      "loss_fn": losses.softmax_loss,
+      "expected_value": 0.
+  }, {
       "loss_fn":
-          losses.softmax_loss,
+          losses.listmle_loss,
+      "expected_value":
+          -sum([
+              log(exp(0.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))),
+              log(exp(3.) / (exp(3.) + exp(1.) + exp(2.))),
+              log(exp(1.) / (exp(1.) + exp(2.))),
+              log(exp(2.) / (exp(2.))),
+          ])
+  }, {
+      "loss_fn":
+          losses.poly1_softmax_loss,
+      "expected_value":
+          1. - sum([
+              0.25 * (exp(0.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))),
+              0.25 * (exp(3.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))),
+              0.25 * (exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))),
+              0.25 * (exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))),
+          ])
+  }, {
+      "loss_fn": losses.unique_softmax_loss,
       "expected_value": 0.
   }, {
       "loss_fn": losses.pairwise_hinge_loss,
@@ -149,11 +219,32 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     scores = jnp.asarray([0., 3., 1., 2.])
     labels = jnp.asarray([0., 0., 0., 0.])
 
-    loss = loss_fn(scores, labels)
+    loss = loss_fn(scores, labels, reduce_fn=jnp.sum)
 
-    self.assertArraysAllClose(jnp.asarray(expected_value), loss)
+    np.testing.assert_allclose(jnp.asarray(expected_value), loss)
 
   @parameterized.parameters([{
+      "loss_fn":
+          losses.softmax_loss,
+      "expected_value":
+          -(2. * log(exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))))
+  }, {
+      "loss_fn":
+          losses.poly1_softmax_loss,
+      "expected_value":
+          -(2. * log(exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))) +
+          (1. - (2. / 3. * exp(2.) /
+                 (exp(0.) + exp(3.) + exp(1.) + exp(2.)) + 1. / 3. *
+                 (exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))))
+  }, {
+      "loss_fn":
+          losses.unique_softmax_loss,
+      "expected_value":
+          -(2. * log(exp(2.) / (exp(0.) + exp(3.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.))))
+  }, {
       "loss_fn": losses.pairwise_hinge_loss,
       "expected_value": 7.
   }, {
@@ -184,9 +275,9 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     labels = jnp.asarray([0., 0., 1., 1.])
     weights = jnp.asarray([1., 1., 2., 1.])
 
-    loss = loss_fn(scores, labels, weights=weights)
+    loss = loss_fn(scores, labels, weights=weights, reduce_fn=jnp.sum)
 
-    self.assertArraysAllClose(jnp.asarray(expected_value), loss)
+    np.testing.assert_allclose(jnp.asarray(expected_value), loss)
 
   @parameterized.parameters([{
       "loss_fn":
@@ -197,6 +288,44 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
           -(2. * log(exp(3.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.))) +
             log(exp(4.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.))))
       ]
+  }, {
+      "loss_fn":
+          losses.listmle_loss,
+      "expected_value": [
+          -sum([
+              log(exp(1.) / (exp(1.) + exp(2.) + exp(0.) + exp(3.))),
+              log(exp(2.) / (exp(2.) + exp(0.) + exp(3.))),
+              log(exp(0.) / (exp(0.) + exp(3.))),
+              log(exp(3.) / (exp(3.))),
+          ]), -sum([
+              log(exp(3.) / (exp(3.) + exp(4.) + exp(1.) + exp(2.))),
+              log(exp(4.) / (exp(4.) + exp(1.) + exp(2.))),
+              log(exp(1.) / (exp(1.) + exp(2.))),
+              log(exp(2.) / (exp(2.))),
+          ])
+      ]
+  }, {
+      "loss_fn":
+          losses.poly1_softmax_loss,
+      "expected_value": [
+          -(log(exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.)))) +
+          (1. - (0.5 * (exp(2.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))) +
+                 0.5 * (exp(1.) / (exp(0.) + exp(3.) + exp(1.) + exp(2.))))),
+          -(2. * log(exp(3.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.))) +
+            log(exp(4.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.)))) +
+          (1. - (2. / 3. * (exp(3.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.))) +
+                 1. / 3. * (exp(4.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.)))))
+      ]
+  }, {
+      "loss_fn":
+          losses.unique_softmax_loss,
+      "expected_value": [
+          -(log(exp(2.) / (exp(0.) + exp(3.) + exp(2.))) +
+            log(exp(1.) / (exp(0.) + exp(3.) + exp(1.)))),
+          -(3. * log(exp(3.) / (exp(3.) + exp(1.) + exp(4.) + exp(2.))) +
+            log(exp(4.) / (exp(1.) + exp(4.) + exp(2.))))
+      ]
   }, {
       "loss_fn": losses.pairwise_hinge_loss,
       "expected_value": [(3. - 1. + 1.) + (3. - 2. + 1.), (4. - 3. + 1.)]
@@ -243,10 +372,11 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     scores = jnp.asarray([[0., 3., 1., 2.], [3., 1., 4., 2.]])
     labels = jnp.asarray([[0., 0., 1., 1.], [2., 0., 1., 0.]])
 
+    loss_fn = functools.partial(loss_fn, reduce_fn=jnp.sum)
     vmap_loss_fn = jax.vmap(loss_fn, in_axes=(0, 0), out_axes=0)
     loss = vmap_loss_fn(scores, labels)
 
-    self.assertArraysAllClose(jnp.asarray(expected_value), loss)
+    np.testing.assert_allclose(jnp.asarray(expected_value), loss)
 
   @parameterized.parameters([{
       "loss_fn": losses.softmax_loss,
@@ -255,6 +385,36 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
           -log(exp(1.5) / (exp(1.) + exp(0.5) + exp(1.5)))
       ],
       "normalizer": 2.
+  }, {
+      "loss_fn": losses.listmle_loss,
+      "expected_value": [
+          -sum([
+              log(exp(2.) / (exp(2.) + exp(1.) + exp(3.))),
+              log(exp(1.) / (exp(1.) + exp(3.))),
+              log(exp(3.) / (exp(3.))),
+          ]), -sum([
+              log(exp(1.5) / (exp(1.5) + exp(1.) + exp(0.5))),
+              log(exp(1.) / (exp(1.) + exp(0.5))),
+              log(exp(0.5) / (exp(0.5))),
+          ])
+      ],
+      "normalizer": 2.
+  }, {
+      "loss_fn": losses.poly1_softmax_loss,
+      "expected_value": [
+          -log(exp(2.) / (exp(2.) + exp(1.) + exp(3.))) +
+          (1. - (exp(2.) / (exp(2.) + exp(1.) + exp(3.)))),
+          -log(exp(1.5) / (exp(1.) + exp(0.5) + exp(1.5))) +
+          (1. - (exp(1.5) / (exp(1.) + exp(0.5) + exp(1.5))))
+      ],
+      "normalizer": 2.
+  }, {
+      "loss_fn": losses.unique_softmax_loss,
+      "expected_value": [
+          -log(exp(2.) / (exp(2.) + exp(1.) + exp(3.))),
+          -log(exp(1.5) / (exp(1.) + exp(0.5) + exp(1.5)))
+      ],
+      "normalizer": 2.
   }, {
       "loss_fn": losses.pairwise_hinge_loss,
       "expected_value": [2., .5],
@@ -294,12 +454,22 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     mean_loss = loss_fn(scores, labels, reduce_fn=jnp.mean)
     sum_loss = loss_fn(scores, labels, reduce_fn=jnp.sum)
 
-    self.assertArraysAllClose(mean_loss, jnp.sum(expected_value) / normalizer)
-    self.assertArraysAllClose(sum_loss, jnp.sum(expected_value))
+    np.testing.assert_allclose(
+        mean_loss, jnp.sum(expected_value) / normalizer, rtol=1E-5)
+    np.testing.assert_allclose(sum_loss, jnp.sum(expected_value))
 
   @parameterized.parameters([{
       "loss_fn": losses.softmax_loss,
       "expected_shape": (2,)
+  }, {
+      "loss_fn": losses.listmle_loss,
+      "expected_shape": (2,)
+  }, {
+      "loss_fn": losses.poly1_softmax_loss,
+      "expected_shape": (2,)
+  }, {
+      "loss_fn": losses.unique_softmax_loss,
+      "expected_shape": (2,)
   }, {
       "loss_fn": losses.pairwise_hinge_loss,
       "expected_shape": (2, 9)
@@ -327,26 +497,28 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
     self.assertEqual(jnp.sum(none_loss), sum_loss)
 
   @parameterized.parameters([
-      losses.softmax_loss, losses.pairwise_hinge_loss,
+      losses.softmax_loss, losses.listmle_loss, losses.pairwise_hinge_loss,
       losses.pairwise_logistic_loss, losses.pointwise_sigmoid_loss,
-      losses.pointwise_mse_loss, losses.pairwise_mse_loss
+      losses.pointwise_mse_loss, losses.pairwise_mse_loss,
+      losses.poly1_softmax_loss, losses.unique_softmax_loss
   ])
   def test_computes_loss_value_with_where(self, loss_fn):
     scores = jnp.asarray([0., 3., 1., 2.])
-    labels = jnp.asarray([0., 0., 1., 1.])
-    where = jnp.asarray([True, True, False, True])
-    expected_scores = jnp.asarray([0., 3., 2.])
-    expected_labels = jnp.asarray([0., 0., 1.])
+    labels = jnp.asarray([0., 0., 2., 1.])
+    where = jnp.asarray([True, True, True, False])
+    expected_scores = jnp.asarray([0., 3., 1.])
+    expected_labels = jnp.asarray([0., 0., 2.])
 
     loss = loss_fn(scores, labels, where=where)
     expected_loss = loss_fn(expected_scores, expected_labels)
 
-    self.assertArraysAllClose(expected_loss, loss)
+    np.testing.assert_allclose(expected_loss, loss)
 
   @parameterized.parameters([
-      losses.softmax_loss, losses.pairwise_hinge_loss,
+      losses.softmax_loss, losses.listmle_loss, losses.pairwise_hinge_loss,
       losses.pairwise_logistic_loss, losses.pointwise_sigmoid_loss,
-      losses.pointwise_mse_loss, losses.pairwise_mse_loss
+      losses.pointwise_mse_loss, losses.pairwise_mse_loss,
+      losses.poly1_softmax_loss, losses.unique_softmax_loss
   ])
   def test_computes_loss_value_with_all_masked(self, loss_fn):
     scores = jnp.asarray([0., 3., 1., 2.])
@@ -355,12 +527,32 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
 
     loss = loss_fn(scores, labels, where=where)
 
-    self.assertArraysAllClose(jnp.asarray(0.), loss)
+    np.testing.assert_allclose(jnp.asarray(0.), loss, atol=1E-7)
 
   @parameterized.parameters([
-      losses.softmax_loss, losses.pairwise_hinge_loss,
+      losses.softmax_loss, losses.listmle_loss, losses.pairwise_hinge_loss,
       losses.pairwise_logistic_loss, losses.pointwise_sigmoid_loss,
-      losses.pointwise_mse_loss, losses.pairwise_mse_loss
+      losses.pointwise_mse_loss, losses.pairwise_mse_loss,
+      losses.poly1_softmax_loss, losses.unique_softmax_loss
+  ])
+  def test_computes_loss_with_arbitrary_batch_dimensions(self, loss_fn):
+    scores = jnp.asarray([2., 3., 1.])
+    labels = jnp.asarray([0., 0., 1.])
+    where = jnp.asarray([False, True, True])
+    original_loss = loss_fn(scores, labels, where=where)
+
+    scores = jnp.asarray([[[[2., 3., 1.]]]])
+    labels = jnp.asarray([[[[0., 0., 1.]]]])
+    where = jnp.asarray([[[[False, True, True]]]])
+    batched_loss = loss_fn(scores, labels, where=where)
+
+    np.testing.assert_allclose(original_loss, batched_loss)
+
+  @parameterized.parameters([
+      losses.softmax_loss, losses.listmle_loss, losses.pairwise_hinge_loss,
+      losses.pairwise_logistic_loss, losses.pointwise_sigmoid_loss,
+      losses.pointwise_mse_loss, losses.pairwise_mse_loss,
+      losses.poly1_softmax_loss, losses.unique_softmax_loss
   ])
   def test_grad_does_not_return_nan_for_zero_labels(self, loss_fn):
     scores = jnp.asarray([0., 3., 1., 2.])
@@ -368,12 +560,14 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
 
     grads = jax.grad(loss_fn)(scores, labels, reduce_fn=jnp.mean)
 
-    self.assertArraysEqual(jnp.isnan(grads), jnp.zeros_like(jnp.isnan(grads)))
+    np.testing.assert_array_equal(
+        jnp.isnan(grads), jnp.zeros_like(jnp.isnan(grads)))
 
   @parameterized.parameters([
-      losses.softmax_loss, losses.pairwise_hinge_loss,
+      losses.softmax_loss, losses.listmle_loss, losses.pairwise_hinge_loss,
       losses.pairwise_logistic_loss, losses.pointwise_sigmoid_loss,
-      losses.pointwise_mse_loss, losses.pairwise_mse_loss
+      losses.pointwise_mse_loss, losses.pairwise_mse_loss,
+      losses.poly1_softmax_loss, losses.unique_softmax_loss
   ])
   def test_grad_does_not_return_nan_with_all_masked(self, loss_fn):
     scores = jnp.asarray([0., 3., 1., 2.])
@@ -382,7 +576,8 @@ class LossesTest(jtu.JaxTestCase, parameterized.TestCase):
 
     grads = jax.grad(loss_fn)(scores, labels, where=where, reduce_fn=jnp.mean)
 
-    self.assertArraysEqual(jnp.isnan(grads), jnp.zeros_like(jnp.isnan(grads)))
+    np.testing.assert_array_equal(
+        jnp.isnan(grads), jnp.zeros_like(jnp.isnan(grads)))
 
 
 def load_tests(loader, tests, ignore):
@@ -397,4 +592,4 @@ def load_tests(loader, tests, ignore):
 
 
 if __name__ == "__main__":
-  absltest.main(testLoader=jtu.JaxTestLoader())
+  absltest.main()
